#!/usr/bin/perl -w
#
# Name       : update-qd.pl
# Description: Update the qdinfo.dat file and qd.c source with any changes from
#              Stanfords project summary file, psummary.html
#              The script formally known as update-qd-files.pl
# Author     : Bas Couwenberg <sebastic@xs4all.nl>
# License    : GPL
#

use strict;
use warnings;
use File::Basename;
use Date::Calc;	
use LWP::UserAgent;
use Getopt::Mixed "nextOption";

use constant TIME_OFS => 946684800;	# Linux epoch is 1970, FAH (Cosm) is 2000

# initialise configuration variables
my $self    = 'update-qd.pl';
my $version = '1.5.6';

# get todays date and convert the numerical month to text
my ($year, $m, $day, $hour, $min, $sec) = Date::Calc::Today_and_Now();
my $month   = Date::Calc::Month_to_Text($m);
my $date    = "$day $month $year";

$m    = '0'.$m    if(length($m)    == 1);
$day  = '0'.$day  if(length($day)  == 1);
$hour = '0'.$hour if(length($hour) == 1);
$min  = '0'.$min  if(length($min)  == 1);
$sec  = '0'.$sec  if(length($sec)  == 1);

my $logdate = "$year-$m-$day $hour:$min:$sec";

# epoch 2000 time in hexidecimal
my $hextime = &hextime();
   $hextime = sprintf("%08s", $hextime);

# lists to store the credits per project in
my %credit_of = ();
my %historic_credit_of = ();

# list of Preferred deadlines per project
my %preferred_of = ();
my %historic_preferred_of = ();

# list of kfactor values per project
my %kfactor_of = ();
my %historic_kfactor_of = ();

# project summary data per project
my %summary_of = ();

# update statistics
my %number_of = (  pt_projects         => 0,
                   ph_projects         => 0,
		   pd_projects         => 0,
		   dh_projects         => 0,
		   pk_projects         => 0,
		   kh_projects         => 0,
 		   new_pt_projects     => 0,
 		   new_pd_projects     => 0,
 		   new_pk_projects     => 0,
 		   updated_pt_projects => 0,
 		   updated_pd_projects => 0,
 		   updated_pk_projects => 0
		);

# EMIII emprotz.dat specific statistics
my %em_projects = ( total_projects   => 0,
                    new_projects     => 0,
                    updated_projects => 0
                  );

my %cfg = ( 'psummary-url'    => 'http://fah-web.stanford.edu/psummary.html',
            'psummary-file'   => './psummary.html',
	    'emprotz-file'    => './emprotz.dat',
	    'qdinfo-file'     => './qdinfo.dat',
	    'qd-file'         => './qd.c',
	    'download'        => 1,
	    'dry-run'         => 0,
	    'force'           => 0,
	    'qdinfo-only'     => 0,
	    'emprotz-too'     => 0,

	    'compile'         => 0,
	    'compile-windows' => 1,
	    'compile-linux'   => 1,
	    'gcc-bin'         => '/usr/bin/gcc',
	    'mingw-bin'       => '/usr/bin/i586-mingw32msvc-gcc',
	    'windows-bin'     => './qd.exe',
	    'linux-bin'       => './qd',
		
	    'date'            => $date,
	    'hextime'         => $hextime,
	
	    'self'            => $self,
	    'version'         => $version,
	    'log'             => 0,
	    'log-file'        => './update-qd.log',
	    'log-data'        => '',
	    'verbose'         => 0
	  );

# get command line parameters to update the configuration
Getopt::Mixed::init("url=s
                     psummary=s
		     emprotz=s
		     qdinfo=s
		     qd=s
		     verbose
		     help
		     compile
		     gcc=s
		     mingw=s
		     windows-bin=s
		     linux-bin=s
		     no-windows
		     no-linux
		     no-download
		     dry-run
		     force
		     qdinfo-only
		     emprotz-too
		     log
		     log-file=s
		     u>url
		     p>psummary
		     e>emprotz
		     i>qdinfo
		     q>qd
		     v>verbose
		     h>help
		     c>compile
		     g>gcc
		     m>mingw
		     w>windows-bin
		     l>linux-bin
		     nw>no-windows
		     nl>no-linux
		     nd>no-download
		     d>dry-run
		     f>force
		     io>qdinfo-only
		     em>emprotz-too
		     L>log
		     Lf>log-file
		    ");

while (my ($option, $value) = nextOption()) {
        if($option eq 'help') {
		print_usage();
		exit;
	}
	elsif($option eq 'url') {
		$cfg{'psummary-url'}    = $value;
	}
	elsif($option eq 'psummary') {
		$cfg{'psummary-file'}   = $value;
	}
	elsif($option eq 'emprotz') {
		$cfg{'emprotz-file'}    = $value;
	}
	elsif($option eq 'qdinfo') {
		$cfg{'qdinfo-file'}     = $value;
	}
	elsif($option eq 'qd') {
		$cfg{'qd-file'}         = $value;
	}
	elsif($option eq 'verbose') {
		$cfg{'verbose'}         = 1;
	}
	elsif($option eq 'compile') {
		$cfg{'compile'}         = 1;
	}
	elsif($option eq 'gcc') {
		$cfg{'gcc-bin'}         = $value;
	}
	elsif($option eq 'mingw') {
		$cfg{'mingw-bin'}       = $value;
	}
	elsif($option eq 'windows-bin') {
		$cfg{'windows-bin'}     = $value;
	}
	elsif($option eq 'linux-bin') {
		$cfg{'linux-bin'}       = $value;
	}
	elsif($option eq 'no-windows') {
		$cfg{'compile-windows'} = 0;
	}
	elsif($option eq 'no-linux') {
		$cfg{'compile-linux'}   = 0;
	}
	elsif($option eq 'no-download') {
		$cfg{'download'}        = 0;
	}
	elsif($option eq 'dry-run') {
		$cfg{'dry-run'}         = 1;
	}
	elsif($option eq 'force') {
		$cfg{'force'}           = 1;
	}
	elsif($option eq 'qdinfo-only') {
		$cfg{'qdinfo-only'}     = 1;
	}
	elsif($option eq 'emprotz-too') {
		$cfg{'emprotz-too'}     = 1;
	}
	elsif($option eq 'log') {
		$cfg{'log'}             = 1;
	}
	elsif($option eq 'log-file') {
		$cfg{'log-file'}        = $value;;
	}
}

# Step 0
# check the required permissions of the files
if(!&valid_permissions()) {
	print "\nErrors encountered.\n";
	
	&abort();
}

# Step 1
# download the psummary.html from Stanford
if(!&download_psummary()) {
	# abort further processing
	# we couldn't download the psummary file
	
	&abort();
}

# Step 2
# read the project data from psummary.html
&parse_psummary();

# Step 3
# read the project/credit data from qdinfo.dat (and optionally from emprotz.dat)
&parse_qdinfo();
&parse_emprotz() if($cfg{'emprotz-too'} == 1);

# Step 4
# update the project/credit data with any changes from psummary.html
&update_qdinfo_projects();
&update_emprotz_projects() if($cfg{'emprotz-too'} == 1);

# if there are no new or updated projects
# we don't have to update qd.c and qdinfo.dat and can quit the program
# we do need to update if we are forced to
if($number_of{new_pt_projects} == 0 && $number_of{updated_pt_projects} == 0 && 
   $number_of{new_pd_projects} == 0 && $number_of{updated_pd_projects} == 0 &&
   $number_of{new_pk_projects} == 0 && $number_of{updated_pk_projects} == 0 &&
   $em_projects{new_projects}  == 0 && $em_projects{updated_projects}  == 0 &&
   !$cfg{'force'}
) {
	if($cfg{'verbose'}) {
		print "No data has changed, no need to update qdinfo.dat and qd.c.\n";

		&print_update_stats();
	}
	
	exit;
}
else {
	# if there were changes, log them if --log
	if($cfg{'log'}) {
		open(F, ">>", $cfg{'log-file'}) || die "Cannot open file: $cfg{'log-file'} ($!)";
		print F $cfg{'log-data'};
		close F;
	}
}

# Step 5
# generate updated qd.c and qdinfo.dat files
# skip qd.c and its possible compile if qdinfo-only is requested
# also update emprotz.dat if emprotz-too is requested

my $qdinfo_data  = &generate_qdinfo();
my $qd_data      = &generate_qd()      if($cfg{'qdinfo-only'} != 1);
my $emprotz_data = &generate_emprotz() if($cfg{'emprotz-too'} == 1);

# Step 6
# Save the generated files

if($cfg{'dry-run'}) {
	# if we are to leave the orginal files alone 
	# overwrite the paths to qdinfo.dat and qd.c,
	# and the paths to the qd.exe and qd binaries

	$cfg{'emprotz-file'} = './new-emprotz.dat';
	$cfg{'qdinfo-file'}  = './new-qdinfo.dat';
	$cfg{'qd-file'}      = './new-qd.c';

	$cfg{'windows-bin'}  = './new-qd.exe';
	$cfg{'linux-bin'}    = './new-qd';
}

if($number_of{new_pt_projects} != 0 || $number_of{updated_pt_projects} != 0 || 
   $number_of{new_pd_projects} != 0 || $number_of{updated_pd_projects} != 0 ||
   $number_of{new_pk_projects} != 0 || $number_of{updated_pk_projects} != 0 ||
   $cfg{'force'}
) {
	print "\nSaving updated qdinfo.dat to: $cfg{'qdinfo-file'}\n";

	open(F, ">", $cfg{'qdinfo-file'}) || die "Cannot open file: $cfg{'qdinfo-file'} ($!)";
	print F $qdinfo_data;
	close F;

	if($cfg{'qdinfo-only'} != 1) {
		print "\nSaving updated qd.c to: $cfg{'qd-file'}\n";
	
		open(F, ">", $cfg{'qd-file'}) || die "Cannot open file: $cfg{'qd-file'} ($!)";
		print F $qd_data;
		close F;
	
		# Step 6 (Optional)
		# Compile qd binary from qd
		if($cfg{'compile'} == 1) {
			&compile_qd();
		}
	}
}
if( $cfg{'emprotz-too'} == 1 &&
   ($em_projects{new_projects} != 0 || $em_projects{updated_projects} != 0)
) {
	print "\nSaving updated emprotz.dat to: $cfg{'emprotz-file'}\n";
	
	open(F, ">", $cfg{'emprotz-file'}) || die "Cannot open file: $cfg{'emprotz-file'} ($!)";
	print F $emprotz_data;
	close F;
}

&print_update_stats();

exit 0;

################################[ Subroutines ]################################

sub abort {
	print "\nProgram aborted!\n";

	exit 1;
}

sub compile_qd {
	my @compile = ();

	if($cfg{'compile-linux'} == 1) {
		my @cmd = ($cfg{'gcc-bin'}, '-Wall', '-m32', '-lm', '-D_GNU_SOURCE', '-DSYSTYPE=0', '-s', '-O2', '-o', $cfg{'linux-bin'}, $cfg{'qd-file'});

		my %command = ( os  => 'Linux',
		                cmd => \@cmd,
			      );

		push @compile, \%command;
	}
	if($cfg{'compile-windows'} == 1) {
		my @cmd = ($cfg{'mingw-bin'}, '-Wall', '-m32', '-lm', '-D_GNU_SOURCE', '-DSYSTYPE=1', '-s', '-O2', '-o', $cfg{'windows-bin'}, $cfg{'qd-file'});

		my %command = ( os  => 'Windows',
				cmd => \@cmd,
			      );
		
		push @compile, \%command;
	}

	foreach my $command (@compile) {
		my $os  = $command->{os};
		my @cmd = @{$command->{cmd}};

		print "\n";
		print "Compiling qd binary for $os ... \n";
		print "Command: @cmd\n";
	
		if(system(@cmd) != 0) {
			print "Command failed: @cmd ($?)\n";
			if ($? == -1) {
				print "Failed to execute: $!\n";
			}
			elsif ($? & 127) {
				printf "Child died with signal %d, %s coredump\n",
					($? & 127),  ($? & 128) ? 'with' : 'without';
			}
			else {
				printf "Child exited with value %d\n", $? >> 8;
			}
	
			exit 1;
		}
	}
}

sub print_update_stats {
	my %longest = ( total => 0,
		        diff  => 0,
		      );

	my @total = (
		      $number_of{pt_projects},
		      $number_of{ph_projects},
		      $number_of{pd_projects},
		      $number_of{dh_projects},
		      $number_of{pk_projects},
		      $number_of{kh_projects},
		    );
	push @total, $em_projects{total_projects} if($cfg{'emprotz-too'} == 1);

	foreach(@total) {
		my $length = length($_);
		$longest{total} = $length if($length > $longest{total});
	}

	my @diff = (
		     $number_of{new_pt_projects},
		     $number_of{new_pd_projects},
		     $number_of{new_pk_projects},
		     $number_of{updated_pt_projects},
		     $number_of{updated_pd_projects},
		     $number_of{updated_pk_projects},
		   );
	push @diff, $em_projects{new_projects} if($cfg{'emprotz-too'} == 1);

	foreach(@diff) {
		my $length = length($_);
		$longest{diff} = $length if($length > $longest{diff});
	}

	print "\n";
	print "Number of projects in qdinfo.dat            : ";
	printf("%$longest{total}s", $number_of{pt_projects});
	print " (+ ";
	printf("%$longest{diff}s", $number_of{new_pt_projects});
	print ")\n";
	
	print "Number of point history entries             : ";
	printf("%$longest{total}s", $number_of{ph_projects});
	print " (+ ";
	printf("%$longest{diff}s", $number_of{updated_pt_projects});
	print ")\n";
	
	print "Number of preferred deadline entries        : ";
	printf("%$longest{total}s", $number_of{pd_projects});
	print " (+ ";
	printf("%$longest{diff}s", $number_of{new_pd_projects});
	print ")\n";

	print "Number of preferred deadline history entries: ";
	printf("%$longest{total}s", $number_of{dh_projects});
	print " (+ ";
	printf("%$longest{diff}s", $number_of{updated_pd_projects});
	print ")\n";

	print "Number of kfactor entries                   : ";
	printf("%$longest{total}s", $number_of{pk_projects});
	print " (+ ";
	printf("%$longest{diff}s", $number_of{new_pk_projects});
	print ")\n";

	print "Number of kfactor history entries           : ";
	printf("%$longest{total}s", $number_of{kh_projects});
	print " (+ ";
	printf("%$longest{diff}s", $number_of{updated_pk_projects});
	print ")\n";

	if($cfg{'emprotz-too'} == 1) {
		print "Number of projects in emprotz.dat           : ";
		printf("%$longest{total}s", $em_projects{total_projects});
		print " (+ ";
		printf("%$longest{diff}s", $em_projects{new_projects});
		print ")\n";
	}
}

sub generate_qd {
	my $qd_head = '';
	my $qd_pt   = '';
	my $qd_pd   = '';
	my $qd_pk   = '';
	my $qd_tail = '';

	my $parse   = 0;
	my $skip_nl = 0;
	open(F, $cfg{'qd-file'}) || die "Cannot open file: $cfg{'qd-file'} ($!)";
	while(<F>) {
		if($parse == 0 && $qd_pt eq '' && $_ !~ /\/\* The following text is automatically generated from \<qdinfo\.dat\> \*\//) {
			$qd_head .= $_;
		}
		elsif(($parse == 1 && $_ !~ /};/) || /\/\* The following text is automatically generated from \<qdinfo\.dat\> \*\//) {
			$parse    = 1;
			$qd_pt   .= $_;
		}
		elsif($parse == 1 && $_ =~ /};/) {
			$parse    = 0;
			$qd_pt   .= $_;
			$skip_nl  = 1;
		}
		elsif($parse == 0 && $qd_pd eq '' && /^struct pdent pdtbl/) {
			$parse    = 2;
			$qd_pd   .= $_;
		}
		elsif($parse == 2 && $_ !~ /};/) {
			$qd_pd   .= $_;
		}
		elsif($parse == 2 && $_ =~ /};/) {
			$parse    = 0;
			$qd_pd   .= $_;
			$skip_nl  = 1;
		}
		elsif($parse == 0 && $qd_pk eq '' && /^struct pkent pktbl/) {
			$parse    = 3;
			$qd_pk   .= $_;
		}
		elsif($parse == 3 && $_ !~ /};/) {
			$qd_pk   .= $_;
		}
		elsif($parse == 3 && $_ =~ /};/) {
			$parse    = 0;
			$qd_pk   .= $_;
		}
		elsif($skip_nl == 1 && $_ =~ /^\n$/) {
			$skip_nl  = 0;
		}
		else {
			$qd_tail .= $_;
		}
	}
	close F;

	$qd_pt  = "/* The following text is automatically generated from <qdinfo.dat> */\n\n";
	$qd_pt .= "#define DATE \"$cfg{'date'}\"\n";
	$qd_pt .= "#define FREV \"$cfg{'functional-revision'}\"\n";
	$qd_pt .= "#define PGEN 0x$cfg{'hextime'}\n";
	$qd_pt .= "struct ptent ptbl[] =\t\t/* Built-in table, from \"pt\" and \"ph\" entries */\n";
	   
	my $i = 0;
	my $p = 0;
	foreach my $project (sort { $a <=> $b } keys %credit_of) {
		# example line:
		#         {  127,  200 },{  128,  200 },{  129,  200 },{  131,  140 },{  132,  280 },
		
		if($i == 0) {
			# first line in the structure, needs opening curly brace
			$qd_pt .= " {\t";
		}
	
		$qd_pt .= "{ ";
		$qd_pt .= sprintf("%5s", $project);
		$qd_pt .= ",";
		$qd_pt .= sprintf("%7s", $credit_of{$project});
		$qd_pt .= " },";
		
		if($p == 4) {
			$qd_pt .= "\n\t";
			$p = -1;
		}
		
		$i++;
		$p++;
	}
	$qd_pt .= "\n\t" if($qd_pt !~ /\n\t$/);
	
	$p = 0;
	foreach my $project (sort { $a <=> $b } keys %historic_credit_of) {
		# example line:
		#         {  171,  200, 0x0A7A948C },{  180,  560, 0x0818538C },{  182,  560, 0x0818538C },
		
		foreach my $project_history (@{$historic_credit_of{$project}}) {
			my ($credits, $hextime) = split /:/, $project_history;
			
			$qd_pt .= "{ ";
			$qd_pt .= sprintf("%5s", $project);
			$qd_pt .= ",";
			$qd_pt .= sprintf("%7s", $credits);
			$qd_pt .= ", 0x$hextime },";
			
			if($p == 2) {
				$qd_pt .= "\n\t";
				$p = -1;
			}
			
			$p++;
		}
	}
	
	$qd_pt .= "\n\t" if($qd_pt !~ /\n\t$/);
	$qd_pt .= "{ 0 }\n";
	$qd_pt .= " };\n";

	# preferred deadline per project list
	$qd_pd  = "\nstruct pdent pdtbl[] =\t\t/* Built-in table, from \"pd\" and \"dh\" entries */\n";
	
	$i = 0;
	$p = 0;
	foreach my $project (sort { $a <=> $b } keys %preferred_of) {
		# example line:
		#         {  964,  259200 },{ 1136,  950400 },{ 1151,  432000 },{ 1155,  432000 },
		
		if($i == 0) {
			# first line in the structure, needs opening curly brace
			$qd_pd .= " {\t";
		}
	
		$qd_pd .= "{ ";
		$qd_pd .= sprintf("%5s", $project);
		$qd_pd .= ",";
		$qd_pd .= sprintf("%8s", $preferred_of{$project});
		$qd_pd .= " },";
		
		if($p == 3) {
			$qd_pd .= "\n\t";
			$p = -1;
		}
		
		$i++;
		$p++;
	}
	$qd_pd .= "\n\t" if($qd_pd !~ /\n\t$/);

	$p = 0;
	foreach my $project (sort { $a <=> $b } keys %historic_preferred_of) {
		# example line:
		#         {  171,  200, 0x0A7A948C },{  180,  560, 0x0818538C },{  182,  560, 0x0818538C },
		
		foreach my $project_history (@{$historic_preferred_of{$project}}) {
			my ($preferred, $hextime) = split /:/, $project_history;
			
			$qd_pd .= "{ ";
			$qd_pd .= sprintf("%5s", $project);
			$qd_pd .= ",";
			$qd_pd .= sprintf("%8s", $preferred);
			$qd_pd .= ", 0x$hextime },";
			
			if($p == 2) {
				$qd_pd .= "\n\t";
				$p = -1;
			}
			
			$p++;
		}
	}

	$qd_pd .= "{ 0 }\n";
	$qd_pd .= " };\n";

	# preferred deadline per project list
	$qd_pk  = "\nstruct pkent pktbl[] =\t\t/* Built-in table, from \"pk\" and \"kh\" entries */\n";
	
	$i = 0;
	$p = 0;
	foreach my $project (sort { $a <=> $b } keys %kfactor_of) {
		# example line:
		#         {  964,  259200 },{ 1136,  950400 },{ 1151,  432000 },{ 1155,  432000 },
		
		if($i == 0) {
			# first line in the structure, needs opening curly brace
			$qd_pk .= " {\t";
		}
	
		$qd_pk .= "{ ";
		$qd_pk .= sprintf("%5s", $project);
		$qd_pk .= ",";
		$qd_pk .= sprintf("%3s", $kfactor_of{$project});
		$qd_pk .= " },";
		
		if($p == 4) {
			$qd_pk .= "\n\t";
			$p = -1;
		}
		
		$i++;
		$p++;
	}
	$qd_pk .= "\n\t" if($qd_pk !~ /\n\t$/);

	$p = 0;
	foreach my $project (sort { $a <=> $b } keys %historic_kfactor_of) {
		# example line:
		#         {  171,  200, 0x0A7A948C },{  180,  560, 0x0818538C },{  182,  560, 0x0818538C },
		
		foreach my $project_history (@{$historic_kfactor_of{$project}}) {
			my ($kfactor, $hextime) = split /:/, $project_history;
			
			$qd_pk .= "{ ";
			$qd_pk .= sprintf("%5s", $project);
			$qd_pk .= ",";
			$qd_pk .= sprintf("%3s", $kfactor);
			$qd_pk .= ", 0x$hextime },";
			
			if($p == 2) {
				$qd_pk .= "\n\t";
				$p = -1;
			}
			
			$p++;
		}
	}

	$qd_pk .= "{ 0 }\n";
	$qd_pk .= " };\n";

	my $qd_data = $qd_head.$qd_pt.$qd_pd.$qd_pk.$qd_tail;

	return $qd_data;
}

sub generate_qdinfo {
	# define the components of the new qdinfo.dat
	# the qdinfo.dat file has 3 sections:
	# - the header        with the functional revision, the date it was generated
	#                     both as human readable date as well as epoch 2000
	#                     timestamp, and by whom the file was generated
	# - the point table   with the current credits per project
	# - the point history with the credit revisions per project

	my $qdinfo_head       = "# <qdinfo.dat>\n\n";
	   $qdinfo_head      .= "da $cfg{'date'} ($cfg{'self'})\n\n";
	   $qdinfo_head      .= "# Latest functional revision number of \"qd\"\n\n";
	   $qdinfo_head      .= "fr $cfg{'functional-revision'}\n\n";
	   $qdinfo_head      .= "# Point table generation time, epoch 2000\n\n";
	   $qdinfo_head      .= "pg $cfg{'hextime'}\n";
	my $point_table       = "\n# Point table entries (points per WU (project, 100 * points))\n\n";
	my $point_history     = "\n# Point history entries (project, 100 * old points, change date epoch 2000)\n\n";
	my $preferred_table   = "\n# Preferred deadline entries (project, deadline in seconds)\n\n";
	my $preferred_history = "\n# Preferred deadline history entries (project, deadline in seconds, change date epoch 2000)\n\n";
	my $kfactor_table     = "\n# Bonus factor (kfactor) entries (project, 100 * kfactor)\n\n";
	my $kfactor_history   = "\n# Bonus factor (kfactor) history entries (project, 100 * kfactor, change date epoch 2000)\n\n";

	# recreate point table with updated project:value combinations
	# each row starts with pt and should have 5 project:value combinations
	my $i = 0;
	my $j = 0;
	foreach my $project (sort { $a <=> $b } keys %credit_of) {
		# example pt line:
		# pt  101  7400   111   500   112   500   113   500   114   500
	
		if($j == 0) {
			$point_table .= "pt ";
		}
		if($j <= 4) {
			$point_table .= sprintf("%5s", $project);
			$point_table .= " ";
			$point_table .= sprintf("%7s", $credit_of{$project});
			$point_table .= "  " if($j < 4);
		}
		if($j == 4) {
			$point_table .= "\n";
			$j = -1;
		}
		
		$i++;
		$j++;
	}
	# last row possible does not contain 5 projects
	if($point_table =~ s/  $/\n/) {}
	if($i == 0) {	# Don't display the header if we have no project data
		$point_table = '';
	}
	
	# recreate point history table with updated credit revisions
	# each row starts with ph and should have a max of 3 credit revisions per line
	# if a project has more than 3 credit revisions addional ph lines need to be
	# written to list them all
	
	$i = 0;
	foreach my $project (sort { $a <=> $b } keys %historic_credit_of) {
		# example ph line:
		# ph 2019 20800 0A77D56C  2019 20100 0A72734C  2019 11600 0A71E6AC
		
		$point_history .= "ph ";
		$j = 0;
		foreach my $project_history (@{$historic_credit_of{$project}}) {
			my ($credits, $hextime) = split /:/, $project_history;
			
			# if this is not the first revision written on the line
			# add 2 spaces as a separator
			if($j > 0) {
				$point_history .= "  ";
			}
	
			$point_history .= sprintf("%5s", $project);
			$point_history .= " ";
			$point_history .= sprintf("%7s", $credits);
			$point_history .= " ";
			$point_history .= sprintf("%08s", $hextime);
			
			if($j == 2) {
				# wrote 3 revistion on one line
				# maximum reached
				
				$point_history .= "\n";
				
				if($#{$historic_credit_of{$project}} > 2) {
					# if this project has more than 3 revisions
					# write a new point history line
					
					$point_history .= "ph ";
				}
				
				$j = -1; # reset the field counter
			}
	
			$j++;
		}
		# add a newline if we didn't write a full line with 3 revisions
		$point_history .= "\n" unless($point_history =~ /\n$/);

		$i++;
		$j++;
	}
	if($i == 0) {	# Don't display the header if we have no project data
		$point_history = '';
	}

	# recreate preferred deadlines with updated project:value combinations
	# each row starts with pd and should have 5 project:value combinations
	$i = 0;
	$j = 0;
	foreach my $project (sort { $a <=> $b } keys %preferred_of) {
		# example pd line:
		# pd  772 25920000   773 25920000   962    86400   963    86400
	
		if($j == 0) {
			$preferred_table .= "pd ";
		}
		if($j <= 3) {
			$preferred_table .= sprintf("%5s", $project);
			$preferred_table .= " ";
			$preferred_table .= sprintf("%8s", $preferred_of{$project});
			$preferred_table .= "  " if($j < 3);
		}
		if($j == 3) {
			$preferred_table .= "\n";
			$j = -1;
		}

		$i++;
		$j++;
	}
	# last row possible does not contain 5 projects
	if($preferred_table =~ s/  $/\n/) {}
	if($i == 0) {	# Don't display the header if we have no project data
		$preferred_table = '';
	}
	
	# recreate preferred deadline history table with updated preferred deadline revisions
	# each row starts with dh and should have a max of 3 preferred deadline revisions per line
	# if a project has more than 3 preferred deadline revisions addional dh lines need to be
	# written to list them all

	$i = 0;	
	foreach my $project (sort { $a <=> $b } keys %historic_preferred_of) {
		# example dh line:
		# dh 2019 20800 0A77D56C  2019 20100 0A72734C  2019 11600 0A71E6AC
		
		$preferred_history .= "dh ";
		$j = 0;
		foreach my $project_history (@{$historic_preferred_of{$project}}) {
			my ($preferred, $hextime) = split /:/, $project_history;
			
			# if this is not the first revision written on the line
			# add 2 spaces as a separator
			if($j > 0) {
				$preferred_history .= "  ";
			}
	
			$preferred_history .= sprintf("%5s", $project);
			$preferred_history .= " ";
			$preferred_history .= sprintf("%8s", $preferred);
			$preferred_history .= " ";
			$preferred_history .= sprintf("%08s", $hextime);
			
			if($j == 2) {
				# wrote 3 revistion on one line
				# maximum reached
				
				$preferred_history .= "\n";
				
				if($#{$historic_preferred_of{$project}} > 2) {
					# if this project has more than 3 revisions
					# write a new preferred deadline history line
					
					$preferred_history .= "dh ";
				}
				
				$j = -1; # reset the field counter
			}
	
			$j++;
		}
		# add a newline if we didn't write a full line with 3 revisions
		$preferred_history .= "\n" unless($preferred_history =~ /\n$/);

		$i++;
	}
	if($i == 0) {	# Don't display the header if we have no project data
		$preferred_history = '';
	}

	# recreate kfactor table with updated project:value combinations
	# each row starts with pk and should have 5 project:value combinations
	$i = 0;
	$j = 0;
	foreach my $project (sort { $a <=> $b } keys %kfactor_of) {
		# example pk line:
		# pd  772 25920000   773 25920000   962    86400   963    86400
	
		if($j == 0) {
			$kfactor_table .= "pk ";
		}
		if($j <= 5) {
			$kfactor_table .= sprintf("%5s", $project);
			$kfactor_table .= " ";
			$kfactor_table .= sprintf("%3s", $kfactor_of{$project});
			$kfactor_table .= "  " if($j < 5);
		}
		if($j == 5) {
			$kfactor_table .= "\n";
			$j = -1;
		}

		$i++;
		$j++;
	}
	# last row possible does not contain 5 projects
	if($kfactor_table =~ s/  $/\n/) {}
	if($i == 0) {	# Don't display the header if we have no project data
		$kfactor_table = '';
	}
	
	# recreate kfactor history table with updated kfactor revisions
	# each row starts with kh and should have a max of 3 kfactor revisions per line
	# if a project has more than 3 kfactor revisions addional kh lines need to be
	# written to list them all

	$i = 0;	
	foreach my $project (sort { $a <=> $b } keys %historic_kfactor_of) {
		# example kh line:
		# kh 2019 20800 0A77D56C  2019 20100 0A72734C  2019 11600 0A71E6AC
		
		$kfactor_history .= "kh ";
		$j = 0;
		foreach my $project_history (@{$historic_kfactor_of{$project}}) {
			my ($kfactor, $hextime) = split /:/, $project_history;
			
			# if this is not the first revision written on the line
			# add 2 spaces as a separator
			if($j > 0) {
				$kfactor_history .= "  ";
			}
	
			$kfactor_history .= sprintf("%5s", $project);
			$kfactor_history .= " ";
			$kfactor_history .= sprintf("%3s", $kfactor);
			$kfactor_history .= " ";
			$kfactor_history .= sprintf("%08s", $hextime);
			
			if($j == 2) {
				# wrote 3 revistion on one line
				# maximum reached
				
				$kfactor_history .= "\n";
				
				if($#{$historic_kfactor_of{$project}} > 2) {
					# if this project has more than 3 revisions
					# write a new kfactor history line
					
					$kfactor_history .= "kh ";
				}
				
				$j = -1; # reset the field counter
			}
	
			$j++;
		}
		# add a newline if we didn't write a full line with 3 revisions
		$kfactor_history .= "\n" unless($kfactor_history =~ /\n$/);

		$i++;
	}
	if($i == 0) {	# Don't display the header if we have no project data
		$kfactor_history = '';
	}

	# compose new qdinfo.dat file
	my $qdinfo_data  = $qdinfo_head.$point_table.$point_history;
	   $qdinfo_data .= $preferred_table.$preferred_history;
	   $qdinfo_data .= $kfactor_table.$kfactor_history;

	return $qdinfo_data;
}

sub generate_emprotz {
	my $emprotz_data = '';

	$emprotz_data .= "\"Version\"\r\n";
	$emprotz_data .= " ". $em_projects{version}{major}    ."\r\n";
	$emprotz_data .= " ". $em_projects{version}{minor}    ."\r\n";
	$emprotz_data .= " ". $em_projects{version}{revision} ."\r\n";

	foreach my $project (sort { $a <=> $b } keys %{$em_projects{project}}) {
		$emprotz_data .= "\"$project\"\r\n";
		$emprotz_data .= " ". $em_projects{project}{$project}{preferred_deadline} ."\r\n";
		$emprotz_data .= " ". $em_projects{project}{$project}{credit}             ."\r\n";
		$emprotz_data .= " ". $em_projects{project}{$project}{frames}             ."\r\n";
	}

	return $emprotz_data;
}


sub update_qdinfo_projects {
	print "\nUpdating qdinfo.dat projects ...\n\n" if($cfg{'verbose'});

	foreach my $project_number (sort keys %summary_of) {
		my $project   = $summary_of{$project_number}{project_number};
		my $preferred = $summary_of{$project_number}{preferred_deadline};
		my $kfactor   = $summary_of{$project_number}{kfactor};
		my $credit    = $summary_of{$project_number}{credit};

		# 2012
		if($project =~ /^\d+$/) {
			# all is well	
		}
		else {
			print "Error: Cannot parse Project Number: $project\n";
			exit 1;	
		}

		# 565.00
		if($credit =~ /^\d+(\.\d+)?$/) {
			$credit  = $credit + 0;       # store \
			$credit *= 100;               # multiply by 100, \
		        	                      # value is saved in qdinfo like that
		
			if($credit_of{$project}) {
				# the project exists in qdinfo.dat
				
				# update the credit value 
				# if it is different from the credit in qdinfo.dat
				if($credit_of{$project} != $credit && $credit_of{$project} ne $credit && $credit ne '' && $credit > 0) {
					my $output  = "Updated project : ";
					   $output .= sprintf("%5s", $project);
					   $output .=  "  new credit*100: ";
					   $output .= sprintf("%7s", $credit);
					   $output .=  "  old credit*100: ";
					   $output .= sprintf("%7s", $credit_of{$project});
					   $output .= "\n";
						
					print $output;
					if($cfg{'log'}) {
						$cfg{'log-data'} .= "[$logdate] ".$output;
					}
					
					# add the old credit value at the beginning 
					# of the revision list of this project
					unshift @{$historic_credit_of{$project}}, "$credit_of{$project}:$cfg{'hextime'}";
					
					# update the amount of credit for this project in the list
					$credit_of{$project} = $credit;
					
					# increase the counter of updated projects
					$number_of{updated_pt_projects} += 1;
				}
			}
			else {
				# the project does not exist in qdinfo.dat yet
				
				my $output  = "    New project : ";
				   $output .= sprintf("%5s", $project);
				   $output .= "  new credit*100: ";
				   $output .= sprintf("%7s", $credit);
				   $output .= "\n";
				
				print $output;
				if($cfg{'log'}) {
					$cfg{'log-data'} .= "[$logdate] ".$output;
				}
				
				# add this project to the list
				$credit_of{$project} = $credit;
				
				# increase the counter of new projects
				$number_of{new_pt_projects} += 1;
			}
		}
		else {
			print "Error: Cannot parse Credit: $credit\n";
			exit 1;	
		}
		
		# 300.00
		if($preferred =~ /^\d+(\.\d+)?$/) {
			# 86400 seconds per day
			$preferred = sprintf("%.0f", $preferred * 86400);

			if($preferred_of{$project}) {
				# the project exists in qdinfo.dat with a preferred deadline
				
				# update the preferred deadline value
				# if it is different from the preferred deadline in qdinfo.dat
				if($preferred_of{$project} != $preferred && $preferred > 0) {
					 my $output  = "Updated preferred deadline for project : ";
					    $output .= sprintf("%5s", $project);
					    $output .=  "  new preferred deadline: ";
					    $output .= sprintf("%8s", $preferred);
					    $output .=  "  old preferred deadline: ";
					    $output .= sprintf("%8s", $preferred_of{$project});
					    $output .= "\n";
					
					print $output;
					if($cfg{'log'}) {
						$cfg{'log-data'} .= "[$logdate] ".$output;
					}
					
					# add the old preferred deadline value at the beginning 
					# of the revision list of this project
					unshift @{$historic_preferred_of{$project}}, "$preferred_of{$project}:$cfg{'hextime'}";

                                	# update the preferred deadline for this project in the list
                                	$preferred_of{$project} = $preferred;

                                	# increase the counter of updated projects
                                	$number_of{updated_pd_projects} += 1;
				}
			}
			else {
				# the project does not exist in qdinfo.dat with a preferred deadline yet
				
				my $output  = "    New preferred deadline for project : ";
				   $output .= sprintf("%5s", $project);
				   $output .= "  new preferred deadline: ";
				   $output .= sprintf("%8s", $preferred);
				   $output .= "\n";
				
				print $output;
				if($cfg{'log'}) {
					$cfg{'log-data'} .= "[$logdate] ".$output;
				}
				
				# add this project to the list
				$preferred_of{$project} = $preferred;
				
				# increase the counter of updated projects
				$number_of{new_pd_projects} += 1;
			}
		}
		else {
			print "Error: Cannot parse Preferred: $preferred\n";
			exit 1;	
		}
		
		# 3.36
		# 0.00
		if($kfactor =~ /^\d+(\.\d+)?$/) {
			# store floating point value as integer in qdinfo.dat
                        $kfactor *= 100;

			if(exists $kfactor_of{$project}) {
				# the project exists in qdinfo.dat with a kfactor
				
				# update the kfactor value
				# if it is different from the kfactor in qdinfo.dat
				if($kfactor_of{$project} != $kfactor && $kfactor_of{$project} ne $kfactor) {
					 my $output  = "Updated kfactor for project : ";
					    $output .= sprintf("%5s", $project);
					    $output .=  "  new kfactor: ";
					    $output .= sprintf("%3s", $kfactor);
					    $output .=  "  old kfactor: ";
					    $output .= sprintf("%3s", $kfactor_of{$project});
					    $output .= "\n";
					
					print $output;
					if($cfg{'log'}) {
						$cfg{'log-data'} .= "[$logdate] ".$output;
					}
					
					# add the old kfactor value at the beginning 
					# of the revision list of this project
					unshift @{$historic_kfactor_of{$project}}, "$kfactor_of{$project}:$cfg{'hextime'}";

                                	# update the kfactor for this project in the list
                                	$kfactor_of{$project} = $kfactor;

                                	# increase the counter of updated projects
                                	$number_of{updated_pk_projects} += 1;
				}
			}
			elsif(!exists $kfactor_of{$project} && $kfactor > 0.00) {
				# the project does not exist in qdinfo.dat with a kfactor yet
				
				my $output  = "    New kfactor for project : ";
				   $output .= sprintf("%5s", $project);
				   $output .= "  new kfactor: ";
				   $output .= sprintf("%3s", $kfactor);
				   $output .= "\n";
				
				print $output;
				if($cfg{'log'}) {
					$cfg{'log-data'} .= "[$logdate] ".$output;
				}
				
				# add this project to the list
				$kfactor_of{$project} = $kfactor;
				
				# increase the counter of updated projects
				$number_of{new_pk_projects} += 1;
			}
		}
		else {
			print "Error: Cannot parse Kfactor: $kfactor\n";
			exit 1;	
		}
	}	
}

sub update_emprotz_projects {
	print "Updating emprotz.dat projects ...\n\n" if($cfg{'verbose'});

	foreach my $project_number (sort keys %summary_of) {
		my $project	       = $summary_of{$project_number}{project_number};
		my $preferred_deadline = $summary_of{$project_number}{preferred_deadline};
		my $credit	       = $summary_of{$project_number}{credit};
		my $frames	       = $summary_of{$project_number}{frames};
		
		if($preferred_deadline =~ /^\d+(\.\d+)?$/) {
			# 86400 seconds per day
			$preferred_deadline = sprintf("%.0f", $preferred_deadline * 86400);
		}
		if($credit =~ /^\d+(\.\d+)?$/) {
			# credits are in whole integers only in EMII data file
			$credit = sprintf("%.0f", $credit + 0);
		}

		if(defined $em_projects{project}{$project}) {
			# the project exists in emprotz.dat
			my $output = '';

			if(
			    ( $em_projects{project}{$project}{preferred_deadline} != $preferred_deadline
			   && $preferred_deadline ne ''
			   && $preferred_deadline > 0
			    )
			   ||
			    ( $em_projects{project}{$project}{credit} != $credit
			   && $credit ne ''
			   && $credit > 0
			    )
			   ||
			    ( $em_projects{project}{$project}{frames} != $frames
			   && $frames ne ''
			   && $frames > 0
			    )
			) {
				# increase the counter of updated projects
				$em_projects{updated_projects} += 1;

				$output .= "Updated project : ";
				$output .= sprintf("%5s", $project);
				$output .= "\n";
			}
			
			# update the preferred deadline value
			# if it is different from the preferred deadline in emprotz.dat
			if($em_projects{project}{$project}{preferred_deadline} != $preferred_deadline && $preferred_deadline ne '' && $preferred_deadline > 0) {
				$output .=  "      new preferred deadline: ";
				$output .= $preferred_deadline;
				$output .=  "      old preferred deadline: ";
				$output .= $em_projects{project}{$project}{preferred_deadline};
				$output .= "\n";
					
				# update the preferred deadline for this project in the list
				$em_projects{project}{$project}{preferred_deadline} = $preferred_deadline;
			}
			# update the credit value 
			# if it is different from the credits in emprotz.dat
			if($em_projects{project}{$project}{credit} != $credit && $credit ne '' && $credit > 0) {
				$output .=  "      new credit: ";
				$output .=  $credit;
				$output .=  "      old credit: ";
				$output .= $em_projects{project}{$project}{credit};
				$output .= "\n";
					
				# update the amount of credits for this project in the list
				$em_projects{project}{$project}{credit} = $credit;
			}
			# update the frames value 
			# if it is different from the frames in emprotz.dat
			if($em_projects{project}{$project}{frames} != $frames && $frames ne '' && $frames > 0) {
				$output .=  "      new frames: ";
				$output .= $frames;
				$output .=  "      old frames: ";
				$output .= $em_projects{project}{$project}{frames};
				$output .= "\n";
					
				# update the number of frames for this project in the list
				$em_projects{project}{$project}{frames} = $frames;
			}

			print $output if($output ne '');
		}
		else {
			# the project does not exist in emprotz.dat yet
			
			my $output  = "    New project : ";
			   $output .= $project;
			   $output .= "\n";
			   $output .= "      new preferred_deadline: ";
			   $output .= $preferred_deadline;
			   $output .= "\n";
			   $output .= "      new credit: ";
			   $output .= $credit;
			   $output .= "\n";
			   $output .= "      new frames: ";
			   $output .= $frames;
			   $output .= "\n";
			
			print $output;
			
			# add this project to the list
			$em_projects{project}{$project}{preferred_deadline} = $preferred_deadline;
			$em_projects{project}{$project}{credit} = $credit;
			$em_projects{project}{$project}{frames} = $frames;
			
			# increase the counter of new projects
			$em_projects{new_projects} += 1;
		}
	}	
}

sub parse_psummary {
	print "\nParsing psummary.html projects ...\n\n" if($cfg{'verbose'});

	my $i = 0;

	open(F, "<", $cfg{'psummary-file'}) || die "Cannot open file: $cfg{'psummary-file'} ($!)";
	while(<F>) {
		$i++;

		# Columns per tablerow:
		#
		#  0 Project Number
		#  1 Server IP
		#  2 Work Unit Name
		#  3 Number of Atoms
		#  4 Preferred (days)
		#  5 Final Deadline (days)
		#  6 Credit
		#  7 Frames
		#  8 Code
		#  9 Description
		# 10 Contact
		# 11 Kfactor
		#	
		# example line to parse:
		#
		#<TR bgcolor="#dfdfdf"> <TD>6013</TD><TD> 130.237.232.140 </TD>
		#<TD>p6013_IBX in water</TD><TD>146808</TD><TD>    3.00</TD>
		#<TD>    3.00</TD><TD>  380.00</TD><TD>100</TD><TD>GRO-A3</TD>
		#<TD><a href=http://fah-web.stanford.edu/cgi-bin/fahproject?p=6013>Description</a></TD>
		#<TD><font size=-1>kasson</font></TD><TD>    3.36</TD></TR>
		#
		# alternate format of line to parse:
		#
		#<TR bgcolor="#dfdfdf"> <TD>5915</TD><TD> 171.64.65.20 </TD>
		#<TD>p5915_proL_ff96_300K</TD><TD>973</TD><TD>    3.00</TD>
		#<TD>    6.00</TD><TD> 1888.00</TD><TD>100</TD><TD>GROGPU2-MT</TD>
		#<TD><a href=http://fah-web.stanford.edu/cgi-bin/fahproject?p=5915>Description</a></TD>
		#<TD><font size=-1>ihaque</font></TD><TD><font size=-1>1</font></TD></TR>
		#
                # Yet another alternate format of line to parse:
                #
                #<TR bgcolor="#dfdfdf"> <TD>6013</TD><TD> 130.237.232.140 </TD>
                #<TD>p6013_IBX in water</TD><TD>146808</TD><TD>3.00</TD><TD>3.00</TD>
                #<TD>380.00</TD><TD>100</TD><TD>GRO-A3</TD>
                #<TD><a href="http://fah-web.stanford.edu/cgi-bin/fahproject?p=6013">Description</a></TD>
                #<TD><font size="-1">kasson</font></TD><TD>3.36</TD></TR>
		#
		# Another format with an email address as contact:
		#
		#<tr bgcolor="#ffffff"> <td>2963</td><td>143.89.28.70</td>
		#<td>p2963</td><td>2963</td><td>10.00</td><td>20.00</td>
		#<td>104.00</td><td>100</td><td>GROMACS</td>
		#<td><a href="http://fah-web.stanford.edu/cgi-bin/fahproject?p=2963">Description</a></td>
		#<td><font size="-1">xuhuihuang@gmail.com</font></td><td>0.00</td></tr>
		#
		# There are multiple description URL formats in use:
		#
		# http://fah-web.stanford.edu/cgi-bin/fahproject?p=1487
		# http://fah-web.stanford.edu/cgi-bin/allprojects#1311
		# http://fah-web.stanford.edu/cgi-bin/fahproject.overusingIPswillbebanned?p=10103
		
		next unless /cgi-bin/;          # only parse project lines
		chomp;                          # remove start/end junk
	
		# Project Number | Server IP | Work Unit Name | Number of Atoms | Preferred (days) |
		# Final Deadline (days) | Credit | Frames | Code | Description | Contact | Kfactor
		if(m!^<TR bgcolor="#\w{6}">\s*<TD>\s*(\d+)\s*</TD><TD>\s*(\d+\.\d+\.\d+\.\d+)\s*</TD><TD>\s*([\w\-\+=\.,: \!\(\)/]+?)\s*</TD><TD>\s*(\d+|)\s*</TD><TD>\s*(\d+\.\d+)\s*</TD><TD>\s*(\d+\.\d+)\s*</TD><TD>\s*(\d+\.\d+)\s*</TD><TD>\s*(\d+)\s*</TD><TD>\s*([\w\-]+)\s*</TD><TD><a href=(?:")?(http://fah-web\.stanford\.edu/cgi-bin/allprojects#\d+|http://fah-web\.stanford\.edu/cgi-bin/fahproject(?:\.overusingIPswillbebanned)?\?p=\d+)(?:")?>Description</a></TD><TD><font size=(?:")?-1(?:")?>([\w@\.]*?)</font></TD><TD>\s*(?:<font size=-1>)?(\d+|\d+\.\d+)(?:</font>)?\s*</TD></TR>\s*$!i) {
			my $project_number     = $1;
			my $server_ip          = $2;
			my $work_unit_name     = $3;
			my $number_of_atoms    = $4;
			my $preferred_deadline = $5;
			my $final_deadline     = $6;
			my $credit             = $7;
			my $frames             = $8;
			my $code               = $9;
			my $description        = $10;
			my $contact	       = $11;
			my $kfactor	       = $12;

			# The contact value may be missing, see:
			# http://foldingforum.org/viewtopic.php?f=18&t=12489&start=30#p123377
                       
			$number_of_atoms = 0	 if(!$number_of_atoms); 
			$contact 	 = 'n/a' if(!$contact);
			$kfactor 	 = 0.00	 if(!$kfactor); 

			my @keys = qw(project_number server_ip work_unit_name number_of_atoms
				      preferred_deadline final_deadline credit frames code
				      description contact kfactor
				      line);

			my %summary = ( project_number     => $project_number,
					server_ip          => $server_ip,
					work_unit_name     => $work_unit_name,
					number_of_atoms    => $number_of_atoms,
					preferred_deadline => $preferred_deadline,
					final_deadline     => $final_deadline,
					credit             => $credit,
					frames             => $frames,
					code               => $code,
					description        => $description,
					contact            => $contact,
					kfactor            => $kfactor,
					line               => $i,
				      );

			if(!$summary_of{$project_number}) {
				$summary_of{$project_number} = \%summary;
			}
			else {
				print "Warning: Multiple records for project: $project_number\n";

				print "Previous:";
				foreach my $key (@keys) {
					print " $summary_of{$project_number}{$key} ";
					print "|" if($key ne 'line');
				}
				print "\n";

				print " Current:";
				foreach my $key (@keys) {
					print " $summary{$key} ";
					print "|" if($key ne 'line');
				}
				print "\n";
				
				print " Columns:";
				foreach my $key (@keys) {
					print " $key ";
					print "|" if($key ne 'line');
				}
				print "\n";

				print "Ignoring current record.\n";
			}
		}
		else {
			# Cannot parse line!
			print "Warning: Cannot parse project line ($i):\n$_\n";
		}
	}
	close F;
}

sub parse_qdinfo {
	# the functional revision is very important
	# if we're unable to get it from qdinfo.dat we should abort
	my $seen_fr = 0;

	open(F, $cfg{'qdinfo-file'}) || die "Cannot open file: $cfg{'qdinfo-file'} ($!)";
	while(<F>) {
		# only parse functional revision, point table and point history
	        next unless /^(fr|pt|ph|pd|dh|pk|kh)/; 
		
		# example funtional revision line:
		# fr 032
		if(/^fr (\d+)/) {
			$cfg{'functional-revision'} = $1;
		
			$seen_fr = 1;
		}
		# example point table lines:
		# pt 2051 16700  2052 16600  2053 13700  2054  6900  2055 12900
		# pt 2056 12800  2057 10600  2059  3800  2060  5700
		elsif(/^pt/) {
			s/^pt\s+//; # strip pt prefix
	
			# The point table lines contain dubble space separated pairs of
			# project and credit data.
			# The first thing on each line are the pt characters to
			# indicate that this line contains point table data.
			# The rest of the line is populated with the actual data, first
			# the project number followed by a space and then the amount of
			# credits for that project multiplied by 100
			#
			# To parse these lines, I remove the pt prefix, and create a
			# list of values by splitting the data on the line on one or
			# more spaces. To keep track of which value in the list is the
			# project number and which is the amount of credits, the
			# variable $i flags the data.
	
			my $i = 0;
			my $project = '';
			foreach(split /\s+/) {
				if($i == 0) {
					# the first value in the pair
					# is the projectnumber
					
					# save the project number
					$project = $_;
	
					# increase the project counter
					$number_of{pt_projects} += 1;
	
					# next value in the list are the credits
					$i = 1;
				}
				else {
					# second value in the pair
					# are the credits of the project
					
					# add the credits to to the list
					# which we will use later to generate
					# the new qdinfo.dat file
	
					$credit_of{$project} = $_;
					
					# next value in the list is the project
					$i = 0;
				}
			}
		}
		# example point history lines:
		# ph 2018 15400 0A584D8C
		# ph 2019 20800 0A77D56C  2019 20100 0A72734C  2019 11600 0A71E6AC
		elsif(/^ph/) {
			# strip carriage return
			if(s/\r//) {}
		
			s/^ph\s+//; # strip ph prefix
	
			# The point history lines contain groups of credit revision 
			# history separted by double spaces. 
			# The first thing on each line are the ph characters to
			# indicate that this line contains point history data.
			# The rest of the line is populated with the actual data, first 
			# the project number followed by a space, then the amount of 
			# credits multiplied by 100, a space, and the epoch 2000 time
			# which indicated when the amount of credits was changed to 
			# that value.
			# Every line contains one or more credit revisions of only one
			# project. Because of the maximum of 80 character per line, 
			# there are at most 3 revision on one line.
			# If a project has more than 3 credit revision, additional 
			# ph lines are created to list all the revisions.
			# So it is very well posible to see a specific project which 
			# has several ph lines.
			# The revisions are ordered from latest up to the oldest. So 
			# the most recent revision is at the beginning of the line.
			#
			# To parse these lines, I remove the ph prefix, and create a
			# list of values by splitting the data on the line on one or 
			# more spaces. To keep track of which value in the list is the 
			# project number, which is the amount of credits and which is 
			# the revision epoch 2000 time, the variable $i flags the data.
			
			my ($project, $credits, $hextime) = ('','','');
			my $i = 0;
			foreach(split /\s+/) {
				if($i == 0) {
					# the first value in the pair
					# is the project number
					
					# save the project number
					$project = $_;
					
					# increase the project counter
					$number_of{ph_projects} += 1;
				}
				elsif($i == 1) {
					# second value in the pair
					# is the credits*100 of the project
					
					# save the amount of credits
					$credits = $_;				
				}
				elsif($i == 2) {
					# third value in the pair
					# is the date of the change in epoch 2000
					# from this machines perspective
					
					# save the epoch 2000 time
					$hextime = $_;
					
					# add the revision data to the list 
					# for this project
					push @{$historic_credit_of{$project}}, "$credits:$hextime";
	
					# reset field counter/flag
					$i = -1; 
				}
				$i++;
			}
		}
		# example preferred deadline lines:
		# pd  772 25920000   773 25920000   962    86400   963    86400
		elsif(/^pd/) {
			s/^pd\s+//; # strip pd prefix
	
			# The preferred deadline lines contain dubble space separated pairs of
			# project and preferred deadline data.
			# The first thing on each line are the pd characters to
			# indicate that this line contains preferred deadline data.
			# The rest of the line is populated with the actual data, first
			# the project number followed by a space and then the number of 
			# seconds in the preferred deadline.
			#
			# To parse these lines, I remove the pd prefix, and create a
			# list of values by splitting the data on the line on one or
			# more spaces. To keep track of which value in the list is the
			# project number and which is the amount of credits, the
			# variable $i flags the data.
	
			my $i = 0;
			my $project = '';
			foreach(split /\s+/) {
				if($i == 0) {
					# the first value in the pair
					# is the projectnumber
					
					# save the project number
					$project = $_;
	
					# increase the project counter
					$number_of{pd_projects} += 1;
	
					# next value in the list are the credits
					$i = 1;
				}
				else {
					# second value in the pair
					# is the preferred deadline of the project
					
					# add the preferred deadline to to the list
					# which we will use later to generate
					# the new qdinfo.dat file
					
					$preferred_of{$project} = $_;
					
					# next value in the list is the project
					$i = 0;
				}
			}
		}
		# example preferred deadline history lines:
		# dh 2018 15400 0A584D8C
		# dh 2019 20800 0A77D56C  2019 20100 0A72734C  2019 11600 0A71E6AC
		elsif(/^dh/) {
			# strip carriage return
			if(s/\r//) {}
		
			s/^dh\s+//; # strip dh prefix
	
			# The preferred deadline history lines contain groups of 
			# preferred deadline revision history separted by double
			# spaces. 
			# The first thing on each line are the dh characters to
			# indicate that this line contains preferred deadline history
			# data.
			# The rest of the line is populated with the actual data, first
			# the project number followed by a space, then preferred deadline
			# time in seconds, a space, and the epoch 2000 time which
			# indicated when the amount of preferred was changed to that value.
			# Every line contains one or more preferred revisions of only one
			# project. Because of the maximum of 80 character per line, 
			# there are at most 3 revisions on one line.
			# If a project has more than 3 preferrred deadline revisions,
			# additional dh lines are created to list all the revisions.
			# So it is very well possible to see a specific project which 
			# has several dh lines.
			# The revisions are ordered from latest up to the oldest. So 
			# the most recent revision is at the beginning of the line.
			#
			# To parse these lines, I remove the dh prefix, and create a
			# list of values by splitting the data on the line on one or 
			# more spaces. To keep track of which value in the list is the 
			# project number, which is the preferred deadline and which is 
			# the revision epoch 2000 time, the variable $i flags the data.
			
			my ($project, $preferred, $hextime) = ('','','');
			my $i = 0;
			foreach(split /\s+/) {
				if($i == 0) {
					# the first value in the pair
					# is the project number
					
					# save the project number
					$project = $_;
					
					# increase the project counter
					$number_of{dh_projects} += 1;
				}
				elsif($i == 1) {
					# second value in the pair
					# is the preferred deadline in seconds of the project
					
					# save the amount of preferred
					$preferred = $_;				
				}
				elsif($i == 2) {
					# third value in the pair
					# is the date of the change in epoch 2000
					# from this machines perspective
					
					# save the epoch 2000 time
					$hextime = $_;
					
					# add the revision data to the list 
					# for this project
					push @{$historic_preferred_of{$project}}, "$preferred:$hextime";
	
					# reset field counter/flag
					$i = -1; 
				}
				$i++;
			}
		}
		# example kfactor lines:
		# pk  772 25920000   773 25920000   962    86400   963    86400
		elsif(/^pk/) {
			s/^pk\s+//; # strip pk prefix
	
			# The kfactor lines contain dubble space separated pairs of
			# project and kfactor data.
			# The first thing on each line are the pk characters to
			# indicate that this line contains kfactor data.
			# The rest of the line is populated with the actual data, first
			# the project number followed by a space and then the kfactor 
			# for that project multiplied by 100
			#
			# To parse these lines, I remove the pk prefix, and create a
			# list of values by splitting the data on the line on one or
			# more spaces. To keep track of which value in the list is the
			# project number and which is the kfactor, the variable $i
			# flags the data.
	
			my $i = 0;
			my $project = '';
			foreach(split /\s+/) {
				if($i == 0) {
					# the first value in the pair
					# is the projectnumber
					
					# save the project number
					$project = $_;
	
					# increase the project counter
					$number_of{pk_projects} += 1;
	
					# next value in the list is the kfactor
					$i = 1;
				}
				else {
					# second value in the pair
					# is the kfactor of the project
					
					# add the kfactor to to the list
					# which we will use later to generate
					# the new qdinfo.dat file
					
					$kfactor_of{$project} = $_;
					
					# next value in the list is the project
					$i = 0;
				}
			}
		}
		# example kfactor history lines:
		# kh 2018 15400 0A584D8C
		# kh 2019 20800 0A77D56C  2019 20100 0A72734C  2019 11600 0A71E6AC
		elsif(/^kh/) {
			# strip carriage return
			if(s/\r//) {}
		
			s/^kh\s+//; # strip kh prefix
	
			# The kfactor history lines contain groups of kfactor
			# revision history separted by double spaces. 
			# The first thing on each line are the kh characters to
			# indicate that this line contains kfactor history data.
			# The rest of the line is populated with the actual data, first
			# the project number followed by a space, then kfactor
			# multiplied by 100, a space, and the epoch 2000 time which
			# indicated when the amount of kfactor was changed to that value.
			# Every line contains one or more kfactor revisions of only one
			# project. Because of the maximum of 80 character per line, 
			# there are at most 3 revisions on one line.
			# If a project has more than 3 preferrred deadline revisions,
			# additional kh lines are created to list all the revisions.
			# So it is very well possible to see a specific project which 
			# has several kh lines.
			# The revisions are ordered from latest up to the oldest. So 
			# the most recent revision is at the beginning of the line.
			#
			# To parse these lines, I remove the kh prefix, and create a
			# list of values by splitting the data on the line on one or 
			# more spaces. To keep track of which value in the list is the 
			# project number, which is the kfactor and which is 
			# the revision epoch 2000 time, the variable $i flags the data.
			
			my ($project, $kfactor, $hextime) = ('','','');
			my $i = 0;
			foreach(split /\s+/) {
				if($i == 0) {
					# the first value in the pair
					# is the project number
					
					# save the project number
					$project = $_;
					
					# increase the project counter
					$number_of{kh_projects} += 1;
				}
				elsif($i == 1) {
					# second value in the pair
					# is the kfactor in seconds of the project
					
					# save the amount of kfactor
					$kfactor = $_;				
				}
				elsif($i == 2) {
					# third value in the pair
					# is the date of the change in epoch 2000
					# from this machines perspective
					
					# save the epoch 2000 time
					$hextime = $_;
					
					# add the revision data to the list 
					# for this project
					push @{$historic_kfactor_of{$project}}, "$kfactor:$hextime";
	
					# reset field counter/flag
					$i = -1; 
				}
				$i++;
			}
		}
	}	
	close F;
	
	if(!$seen_fr) {
		print "Unable to find the functional revision in the file!\n";
		print "Please verify the \"fr\" entry in qdinfo.dat!\n";

		&abort();
	}
}

sub parse_emprotz {
	my $data = '';
	open(F, $cfg{'emprotz-file'}) || die "Cannot open file: $cfg{'emprotz-file'} ($!)";
	while(<F>) {
		$data .= $_;
	}
	close F;

	# "Version"
	#  5
	#  0
	#  6
	# "962"
	#  86400
	#  28
	#  100
	# "963"
	#  86400
	#  23
	#  50

	while($data =~ /"(Version|\d+)"\r\n (\d+)\r\n (\d+)\r\n (\d+)\r\n/) {
		$data = $';

		my $s1 = $1;
		my $s2 = $2;
		my $s3 = $3;
		my $s4 = $4;

		if($s1 eq 'Version') {
			$em_projects{version}{major}    = $s2;
			$em_projects{version}{minor}    = $s3;
			$em_projects{version}{revision} = $s4;
		}
		else {
			$em_projects{project}{$s1}{preferred_deadline} = $s2;
			$em_projects{project}{$s1}{credit}             = $s3;
			$em_projects{project}{$s1}{frames}             = $s4;
			
			$em_projects{total_projects} += 1;
		}
	}
}

sub download_psummary {
	# skip the download if requested
	if(!$cfg{'download'}) {
		if($cfg{'verbose'}) {
			print "Skipping download of URL: $cfg{'psummary-url'}\n";
		}
		
		# check if there is actually a file on disk
		# abort if missing/unreadable
		if(-r $cfg{'psummary-file'}) {
			print "Using file already on disk: $cfg{'psummary-file'}\n" if($cfg{'verbose'});
		}
		else {
			print "Error: Cannot read psummary file! ($cfg{'psummary-file'})\n";
			print "Cannot continue without psummary file!\n";
			
			&abort();
		}
		
		return 1;
	}
		
	print "Downloading URL: $cfg{'psummary-url'} ... " if($cfg{'verbose'});

	# create useragent object which will handle the HTTP request
	my $ua  = LWP::UserAgent->new();
	   $ua->agent("$cfg{self}/$cfg{version}");
	
	# create and send the HTTP request
	my $req = HTTP::Request->new(GET => $cfg{'psummary-url'});
	my $res = $ua->request($req);
	if($res->is_success) {
		print "Success\n" if($cfg{'verbose'});
		
		print "Saving file to disc: $cfg{'psummary-file'}\n" if($cfg{'verbose'});
		open(F, ">", $cfg{'psummary-file'}) || die "Cannot open file: $cfg{'psummary-file'} ($!)";
		print F $res->content;
		close F;

		return 1;
	}
	else {
		print "Failed!\n" if($cfg{'verbose'});
		print "Error: Failed to retrieve URL! ($cfg{'psummary-url'})\n";
		print "Server returned: ".$res->code." ".$res->message."\n";
		
		return undef;
	}
}

sub valid_permissions {
	print "Checking file permissions\n" if($cfg{'verbose'});

	# psummary.html
	# only check the file if it exist
	# check the parent directory if it has yet to be created
	#
	# needs read/write permissions to the file
	# or write permissions to the directory in which it is to be created
	
	print " psummary.html : " if($cfg{'verbose'});
	if(-e $cfg{'psummary-file'}) {
		if(!-w $cfg{'psummary-file'} || !-r $cfg{'psummary-file'}) {
			print "Error!\n" if($cfg{'verbose'});
			print "  file: $cfg{'psummary-file'}\n";
			
			if(!-r $cfg{'psummary-file'}) {
				print "   Cannot read file!\n";
			}
			if(!-w $cfg{'psummary-file'}) {
				print "   Cannot write file!\n";
			}

			return undef;
		}
		else {
			print "OK\n" if($cfg{'verbose'});
		}
	}
	else {
		my $dir = dirname($cfg{'psummary-file'});
		if(!-w $dir) {
			print "Error!\n" if($cfg{'verbose'});
			print "  file: $cfg{'psummary-file'}\n";
			print "   Cannot write directory!\n";
			
			return undef;
		}
		else {
			print "OK\n" if($cfg{'verbose'});
		}
	}

	# emprotz.dat (optional)
	# needs read/write permissions
	if(-e $cfg{'emprotz-file'}) {
		print " emprotz.dat   : " if($cfg{'verbose'});
		if(!-r $cfg{'emprotz-file'} || !-w $cfg{'emprotz-file'}) {
			print "Error!\n" if($cfg{'verbose'});
			print "  file: $cfg{'emprotz-file'}\n";
			if(!-r $cfg{'emprotz-file'}) {
				print "   Cannot read file!\n";
			}
			if(!-w $cfg{'emprotz-file'}) {
				print "   Cannot write file!\n";
			}
			
			return undef;
		}
		else {
			print "OK\n" if($cfg{'verbose'});
		}
	}

	# qdinfo.dat
	# needs read/write permissions
	print " qdinfo.dat    : " if($cfg{'verbose'});
	if(!-r $cfg{'qdinfo-file'} || !-w $cfg{'qdinfo-file'}) {
		print "Error!\n" if($cfg{'verbose'});
		print "  file: $cfg{'qdinfo-file'}\n";
		if(!-r $cfg{'qdinfo-file'}) {
			print "   Cannot read file!\n";
		}
		if(!-w $cfg{'qdinfo-file'}) {
			print "   Cannot write file!\n";
		}
		
		return undef;
	}
	else {
		print "OK\n" if($cfg{'verbose'});
	}
	
	# qd.c
	# needs read/write permissions
	# check only if qdinfo-only is not requested
	if($cfg{'qdinfo-only'} != 1) {
		print " qd.c          : " if($cfg{'verbose'});
		if(!-r $cfg{'qd-file'} || !-w $cfg{'qd-file'}) {
			print "Error!\n" if($cfg{'verbose'});
			print "  file: $cfg{'qd-file'}\n";
			if(!-r $cfg{'qd-file'}) {
				print "   Cannot read file!\n";
			}
			if(!-w $cfg{'qd-file'}) {
				print "   Cannot write file!\n";
			}
		
			return undef;
		}
		else {
			print "OK\n" if($cfg{'verbose'});
		}

		# if we need to compile, check those files and apps too
		if($cfg{'compile'} == 1) {
			if($cfg{'compile-linux'} == 1) {
				# check the linux specific parameters
				
				print " gcc           : " if($cfg{'verbose'});
				if(!-x $cfg{'gcc-bin'}) {
					print "Error!\n" if($cfg{'verbose'});
					
					print "  file: $cfg{'gcc-bin'}\n";
					print "   Cannot execute file!\n";
				}
				else {
					print "OK\n" if($cfg{'verbose'});
				}
	
				print " qd            : " if($cfg{'verbose'});
				if(-e $cfg{'linux-bin'}) {
					if(!-w $cfg{'linux-bin'}) {
						print "Error!\n" if($cfg{'verbose'});
						
						print "  file: $cfg{'linux-bin'}\n";
						print "   Cannot write file!\n";
	
						return undef;
					}
					else {
						print "OK\n" if($cfg{'verbose'});
					}
				}
				else {
					my $dir = dirname($cfg{'linux-bin'});
					if(!-w $dir) {
						print "Error!\n" if($cfg{'verbose'});
						print "  file: $cfg{'linux-bin'}\n";
						print "   Cannot write directory!\n";
				
						return undef;
					}
					else {
						print "OK\n" if($cfg{'verbose'});
					}
				}
			}
			if($cfg{'compile-windows'}) {
				# check the windows specific parameters
				
				print " mingw         : " if($cfg{'verbose'});
				if(!-x $cfg{'mingw-bin'}) {
					print "Error!\n" if($cfg{'verbose'});
					
					print "  file: $cfg{'mingw-bin'}\n";
					print "   Cannot execute file!\n";
				}
				else {
					print "OK\n" if($cfg{'verbose'});
				}
	
				print " qd.exe        : " if($cfg{'verbose'});
				if(-e $cfg{'windows-bin'}) {
					if(!-w $cfg{'windows-bin'}) {
						print "Error!\n" if($cfg{'verbose'});
							
						print "  file: $cfg{'windows-bin'}\n";
						print "   Cannot write file!\n";
	
						return undef;
					}
					else {
						print "OK\n" if($cfg{'verbose'});
					}
				}
				else {
					my $dir = dirname($cfg{'windows-bin'});
					if(!-w $dir) {
						print "Error!\n" if($cfg{'verbose'});
						print "  file: $cfg{'windows-bin'}\n";
						print "   Cannot write directory!\n";
				
						return undef;
					}
					else {
						print "OK\n" if($cfg{'verbose'});
					}
				}
			}
		}
	}

	# update-qd log file
	# needs read/write permissions
	if($cfg{'log'}) {
		print " update-qd.log : " if($cfg{'verbose'});
		if($cfg{'log-file'}) {
			if(-e $cfg{'log-file'}) {
				if(-r $cfg{'log-file'} && -w $cfg{'log-file'}) {
					print "OK\n" if($cfg{'verbose'});
				}
				else {
					print "Error!\n" if($cfg{'verbose'});
					print "  file: $cfg{'log-file'}\n";
					if(!-r $cfg{'log-file'}) {
						print "   Cannot read file!\n";
					}
					if(!-w $cfg{'log-file'}) {
						print "   Cannot write file!\n";
					}
					
					return undef;
				}
			}
			else {
				if(-w dirname($cfg{'log-file'})) {
					print "OK\n" if($cfg{'verbose'});
				}
				else {
					print "Error!\n" if($cfg{'verbose'});
					print "  file: $cfg{'log-file'}\n";
					print "   Cannot write directory!\n";
				}
			}
		}
		else {
			print "Error!\n" if($cfg{'verbose'});
			print "  file: none\n";
			print "   No log-file specified!\n";

			return undef;
		}
	}
	
	print "\n" if($cfg{'verbose'});

	return 1;
}

sub decode_hextime {
	# convert a hexadecimal epoch 2000 value to human readable date string
	
	# Some notes on epoch 2000 vs epoch 1970
	#
	# epoch 2000/Cosm
	# starts: January 1, 2000, 00:00:00 UTC
	#
	# epoch 1970/UNIX
	# starts: January 1, 1970, 00:00:00 UTC
	#
	# difference between the two dates: 946684800 seconds (TIME_OFS)
	#
	# example epoch 2000 strings:
	# 0A8DA04A
	# 0817020C
	#
	# $date_string = scalar localtime (hex("0A8DA04A") + 946684800);
	#
	# add the seconds between epoch 2000 and epoch 1970 to the decimal
	# value of the hex date to get the total amount of seconds since the
	# UNIX epoch convert that to human readable date through scalar
	# localtime
	#
	
	my $hextime = $_[0];

	# scalar localtime hex($_), uses the timezone of this machine
	# not Dick Howells who originally generated the qdinfo.dat files
	
	return scalar localtime (hex($hextime) + TIME_OFS);
}

sub hextime {
	# convert current date to hex
	
	my $time  = time;                 # seconds since epoch 1970
	   $time -= TIME_OFS;             # seconds since epoch 2000
	   $time  = sprintf("%X", $time); # seconds to hex in uppercase
	   
	return $time;
}

sub print_usage {
	$cfg{'verbose'}         = $cfg{'verbose'}         == 1 ? 'true' : 'false';
	$cfg{'download'}        = $cfg{'download'}        == 1 ? 'true' : 'false';
	$cfg{'dry-run'}         = $cfg{'dry-run'}         == 1 ? 'true' : 'false';
	$cfg{'force'}           = $cfg{'force'}           == 1 ? 'true' : 'false';
	$cfg{'qdinfo-only'}     = $cfg{'qdinfo-only'}     == 1 ? 'true' : 'false';
	$cfg{'emprotz-too'}     = $cfg{'emprotz-too'}     == 1 ? 'true' : 'false';
	$cfg{'compile'}         = $cfg{'compile'}         == 1 ? 'true' : 'false';
	$cfg{'compile-linux'}   = $cfg{'compile-linux'}   == 0 ? 'true' : 'false';
	$cfg{'compile-windows'} = $cfg{'compile-windows'} == 0 ? 'true' : 'false';
	$cfg{'log'}             = $cfg{'log'}             == 1 ? 'true' : 'false';
	
	print "$cfg{self} v$cfg{version}\n\n";
	print "Usage: $cfg{self} [OPTIONS]\n\n";
	print "Options:\n";
	
	print "-u, --url <url>           URL to psummary.html\n";
	print "                           default: $cfg{'psummary-url'}\n";
	print "-p, --psummary <file>     Path to save psummary.html to\n";
	print "                           default: $cfg{'psummary-file'}\n";
	print "-q, --qd <file>           Path to qd.c file\n";
	print "                           default: $cfg{'qd-file'}\n";
	print "-i, --qdinfo <file>       Path to qdinfo.dat file\n";
	print "                           default: $cfg{'qdinfo-file'}\n";
	print "-e, --emprotz <file>      Path to emprotz.dat file\n";
	print "                           default: $cfg{'emprotz-file'}\n";
	print "--nd, --no-download       Don't download the psummary.html file\n";
	print "                           default: $cfg{'download'}\n";
	print "--io, --qdinfo-only       Only update qdinfo.dat, skip qd.c\n";
	print "                           default: $cfg{'qdinfo-only'}\n";
	print "--em, --emprotz-too       Also update emprotz.dat\n";
	print "                           default: $cfg{'emprotz-too'}\n";
	print "-f, --force               Force regeneration of point table\n";
	print "                           default: $cfg{'force'}\n";
	print "-d, --dry-run             Don't overwrite the qd.c and qdinfo.dat\n";
	print "                          Will save files to ./new-qd.c and ./new-qdinfo.dat\n";
	print "                          Will compile to ./new-qd.exe and ./new-qd\n";
	print "                           default: $cfg{'dry-run'}\n";
	print "-L, --log                 Log new and/or updated projects.\n";
	print "                           default: $cfg{'log'}\n";
	print "--Lf, --log-file <file>   Path to log file.\n";
	print "                           default: $cfg{'log-file'}\n";
	
	print "\n";	
	print "-v, --verbose             Enable verbose output\n";
	print "                           default: $cfg{'verbose'}\n";
	print "-h, --help                Print usage information\n";
	
	print "\nCompile Options:\n";
	print "-c, --compile             Enable compilation of updated qd.c\n";
	print "                           default: $cfg{'compile'}\n";
	print "-g, --gcc <file>          Path to gcc executable\n";
	print "                           default: $cfg{'gcc-bin'}\n";
	print "-m, --mingw <file>        Path to mingw executable\n";
	print "                           default: $cfg{'mingw-bin'}\n";
	print "-l, --linux-bin <file>    Path to the Linux qd binary\n";
	print "                           default: $cfg{'linux-bin'}\n";
	print "-w, --windows-bin <file>  Path to the Windows qd binary\n";
	print "                           default: $cfg{'windows-bin'}\n";
	print "--nl, --no-linux          Don't compile the Linux binary\n";
	print "                           default: $cfg{'compile-linux'}\n";
	print "--nw, --no-windows        Don't compile the Windows binary\n";
	print "                           default: $cfg{'compile-windows'}\n";
}
